@@ -1,6 +1,6 @@
 
 /*
- * $Id: ftp.c,v 1.280.2.9 1999/05/11 20:34:36 wessels Exp $
+ * $Id: ftp.c,v 1.286.2.19 2000/03/16 06:24:36 wessels Exp $
  *
  * DEBUG: section 9     File Transfer Protocol (FTP)
  * AUTHOR: Harvest Derived
@@ -12,10 +12,10 @@
  *  Internet community.  Development is led by Duane Wessels of the
  *  National Laboratory for Applied Network Research and funded by the
  *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  the Regents of the University of California.  Please see the
+ *  COPYRIGHT file for full details.  Squid incorporates software
+ *  developed and/or copyrighted by other sources.  Please see the
+ *  CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -85,9 +85,11 @@ typedef struct _Ftpdata {
     request_t *request;
     char user[MAX_URL];
     char password[MAX_URL];
+    int password_url;
     char *reply_hdr;
     int reply_hdr_state;
     char *title_url;
+    char *base_href;
     int conn_att;
     int login_att;
     ftp_state_t state;
@@ -163,8 +165,11 @@ static void ftpUnhack(FtpStateData * ftpState);
 static void ftpScheduleReadControlReply(FtpStateData *, int);
 static void ftpHandleControlReply(FtpStateData *);
 static char *ftpHtmlifyListEntry(char *line, FtpStateData * ftpState);
+static void ftpFailed(FtpStateData *, err_type);
+static void ftpFailedErrorMessage(FtpStateData *, err_type);
 
-/* State machine functions
+/*
+ * State machine functions
  * send == state transition
  * read == wait for response, and select next state transition
  * other == Transition logic
@@ -234,24 +239,24 @@ Quit			-
 
 FTPSM *FTP_SM_FUNCS[] =
 {
-    ftpReadWelcome,
-    ftpReadUser,
-    ftpReadPass,
-    ftpReadType,
-    ftpReadMdtm,
-    ftpReadSize,
-    ftpReadPort,
-    ftpReadPasv,
-    ftpReadCwd,
+    ftpReadWelcome,		/* BEGIN */
+    ftpReadUser,		/* SENT_USER */
+    ftpReadPass,		/* SENT_PASS */
+    ftpReadType,		/* SENT_TYPE */
+    ftpReadMdtm,		/* SENT_MDTM */
+    ftpReadSize,		/* SENT_SIZE */
+    ftpReadPort,		/* SENT_PORT */
+    ftpReadPasv,		/* SENT_PASV */
+    ftpReadCwd,			/* SENT_CWD */
     ftpReadList,		/* SENT_LIST */
     ftpReadList,		/* SENT_NLST */
-    ftpReadRest,
-    ftpReadRetr,
-    ftpReadStor,
-    ftpReadQuit,
-    ftpReadTransferDone,
-    ftpSendReply,
-    ftpReadMkdir
+    ftpReadRest,		/* SENT_REST */
+    ftpReadRetr,		/* SENT_RETR */
+    ftpReadStor,		/* SENT_STOR */
+    ftpReadQuit,		/* SENT_QUIT */
+    ftpReadTransferDone,	/* READING_DATA (RETR,LIST,NLST) */
+    ftpSendReply,		/* WRITING_DATA (STOR) */
+    ftpReadMkdir		/* SENT_MKDIR */
 };
 
 static void
@@ -294,6 +299,7 @@ ftpStateFree(int fdnotused, void *data)
     safe_free(ftpState->old_reply);
     safe_free(ftpState->old_filepath);
     safe_free(ftpState->title_url);
+    safe_free(ftpState->base_href);
     safe_free(ftpState->filepath);
     safe_free(ftpState->data.host);
     if (ftpState->data.fd > -1) {
@@ -313,6 +319,7 @@ ftpLoginParser(const char *login, FtpStateData * ftpState, int escaped)
 	xstrncpy(ftpState->password, s + 1, MAX_URL);
 	if (escaped)
 	    rfc1738_unescape(ftpState->password);
+	ftpState->password_url = 1;
     } else {
 	xstrncpy(ftpState->password, null_string, MAX_URL);
     }
@@ -330,18 +337,14 @@ ftpTimeout(int fd, void *data)
     FtpStateData *ftpState = data;
     StoreEntry *entry = ftpState->entry;
     debug(9, 4) ("ftpTimeout: FD %d: '%s'\n", fd, storeUrl(entry));
-    if (entry->store_status == STORE_PENDING) {
-	if (entry->mem_obj->inmem_hi == 0) {
-	    fwdFail(ftpState->fwd,
-		errorCon(ERR_READ_TIMEOUT, HTTP_GATEWAY_TIMEOUT));
-	}
+    if (SENT_PASV == ftpState->state && fd == ftpState->data.fd) {
+	/* stupid ftp.netscape.com */
+	ftpState->fwd->flags.dont_retry = 0;
+	ftpState->fwd->flags.ftp_pasv_failed = 1;
+	debug(9, 1) ("ftpTimeout: timeout in SENT_PASV state\n");
     }
-    if (ftpState->data.fd > -1) {
-	comm_close(ftpState->data.fd);
-	ftpState->data.fd = -1;
-    }
-    comm_close(ftpState->ctrl.fd);
-    /* don't modify ftpState here, it has been freed */
+    ftpFailed(ftpState, ERR_READ_TIMEOUT);
+    /* ftpFailed closes ctrl.fd and frees ftpState */
 }
 
 static void
@@ -361,7 +364,7 @@ ftpListingStart(FtpStateData * ftpState)
     storeAppendPrintf(e, "</TITLE>\n");
     if (ftpState->flags.use_base)
 	storeAppendPrintf(e, "<BASE HREF=\"%s\">\n",
-	    ftpState->title_url);
+	    ftpState->base_href);
     storeAppendPrintf(e, "</HEAD><BODY>\n");
     if (ftpState->cwd_message) {
 	storeAppendPrintf(e, "<PRE>\n");
@@ -419,11 +422,10 @@ ftpListingFinish(FtpStateData * ftpState)
     }
     storeAppendPrintf(e, "<HR>\n");
     storeAppendPrintf(e, "<ADDRESS>\n");
-    storeAppendPrintf(e, "Generated %s by %s (<a href=\"http://squid.nlanr.net/Squid/\">%s</a>)\n",
+    storeAppendPrintf(e, "Generated %s by %s (%s)\n",
 	mkrfc1123(squid_curtime),
 	getMyHostname(),
-	full_appname_string,
-	version_string);
+	full_appname_string);
     storeAppendPrintf(e, "</ADDRESS></BODY></HTML>\n");
     storeBufferFlush(e);
 }
@@ -565,6 +567,8 @@ ftpListParseParts(const char *buf, struct _ftp_flags flags)
 	ct = buf + 1;
 	p->type = 0;
 	while (ct && *ct) {
+	    long lt;
+	    time_t t;
 	    switch (*ct) {
 	    case '\t':
 		sscanf(ct + 1, "%[^,]", sbuf);
@@ -574,8 +578,10 @@ ftpListParseParts(const char *buf, struct _ftp_flags flags)
 		sscanf(ct + 1, "%d", &(p->size));
 		break;
 	    case 'm':
-		sscanf(ct + 1, "%d", &i);
-		p->date = xstrdup(ctime((time_t *) & i));
+		if (1 != sscanf(ct + 1, "%ld", &lt))
+		    break;
+		t = lt;
+		p->date = xstrdup(ctime(&t));
 		*(strstr(p->date, "\n")) = '\0';
 		break;
 	    case '/':
@@ -697,7 +703,7 @@ ftpHtmlifyListEntry(char *line, FtpStateData * ftpState)
 	}
     }
     /* {icon} {text} . . . {date}{size}{chdir}{view}{download}{link}\n  */
-    xstrncpy(href, rfc1738_escape(parts->name), 2048);
+    xstrncpy(href, rfc1738_escape_part(parts->name), 2048);
     xstrncpy(text, parts->showname, 2048);
     switch (parts->type) {
     case 'd':
@@ -831,11 +837,13 @@ ftpReadComplete(FtpStateData * ftpState)
 {
     debug(9, 3) ("ftpReadComplete\n");
     /* Connection closed; retrieval done. */
-    if (ftpState->flags.html_header_sent)
-	ftpListingFinish(ftpState);
-    if (!ftpState->flags.put) {
-	storeTimestampsSet(ftpState->entry);
-	fwdComplete(ftpState->fwd);
+    if (ftpState->data.fd > -1) {
+	/*
+	 * close data socket so it does not occupy resources while
+	 * we wait
+	 */
+	comm_close(ftpState->data.fd);
+	ftpState->data.fd = -1;
     }
     /* expect the "transfer complete" message on the control socket */
     ftpScheduleReadControlReply(ftpState, 1);
@@ -895,8 +903,9 @@ ftpDataRead(int fd, void *data)
 		data,
 		Config.Timeout.read);
 	} else {
-	    assert(mem->inmem_hi > 0);
-	    ftpDataTransferDone(ftpState);
+	    ftpFailed(ftpState, ERR_READ_ERROR);
+	    /* ftpFailed closes ctrl.fd and frees ftpState */
+	    return;
 	}
     } else if (len == 0) {
 	ftpReadComplete(ftpState);
@@ -930,12 +939,10 @@ ftpCheckAuth(FtpStateData * ftpState, const HttpHeader * req_hdr)
     char *orig_user;
     const char *auth;
     ftpLoginParser(ftpState->request->login, ftpState, FTP_LOGIN_ESCAPED);
-    if (ftpState->user[0] && ftpState->password[0])
-	return 1;		/* name and passwd both in URL */
-    if (!ftpState->user[0] && !ftpState->password[0])
-	return 1;		/* no name or passwd */
-    if (ftpState->password[0])
-	return 1;		/* passwd with no name? */
+    if (!ftpState->user[0])
+	return 1;		/* no name */
+    if (ftpState->password_url || ftpState->password[0])
+	return 1;		/* passwd provided in URL */
     /* URL has name, but no passwd */
     if (!(auth = httpHeaderGetAuth(req_hdr, HDR_AUTHORIZATION, "Basic")))
 	return 0;		/* need auth header */
@@ -967,7 +974,6 @@ ftpCheckUrlpath(FtpStateData * ftpState)
     ftpState->flags.use_base = 1;
     /* check for null path */
     if (!l) {
-	stringReset(&request->urlpath, "/");
 	ftpState->flags.isdir = 1;
 	ftpState->flags.root_dir = 1;
     } else if (!strCmp(request->urlpath, "/%2f/")) {
@@ -1005,6 +1011,21 @@ ftpBuildTitleUrl(FtpStateData * ftpState)
     if (request->port != urlDefaultPort(PROTO_FTP))
 	snprintf(&t[strlen(t)], len - strlen(t), ":%d", request->port);
     strcat(t, strBuf(request->urlpath));
+    t = ftpState->base_href = xcalloc(len, 1);
+    strcat(t, "ftp://");
+    if (strcmp(ftpState->user, "anonymous")) {
+	strcat(t, rfc1738_escape_part(ftpState->user));
+	if (ftpState->password_url) {
+	    strcat(t, ":");
+	    strcat(t, rfc1738_escape_part(ftpState->password));
+	}
+	strcat(t, "@");
+    }
+    strcat(t, request->host);
+    if (request->port != urlDefaultPort(PROTO_FTP))
+	snprintf(&t[strlen(t)], len - strlen(t), ":%d", request->port);
+    strcat(t, strBuf(request->urlpath));
+    strcat(t, "/");
 }
 
 void
@@ -1030,7 +1051,12 @@ ftpStart(FwdState * fwd)
     ftpState->data.fd = -1;
     ftpState->size = -1;
     ftpState->mdtm = -1;
-    ftpState->flags.pasv_supported = 1;
+    if (!Config.Ftp.passive)
+	ftpState->flags.rest_supported = 0;
+    else if (fwd->flags.ftp_pasv_failed)
+	ftpState->flags.pasv_supported = 0;
+    else
+	ftpState->flags.pasv_supported = 1;
     ftpState->flags.rest_supported = 1;
     ftpState->fwd = fwd;
     comm_add_close_handler(fd, ftpStateFree, ftpState);
@@ -1073,7 +1099,6 @@ ftpStart(FwdState * fwd)
     ftpState->data.size = SQUID_TCP_SO_RCVBUF;
     ftpState->data.freefunc = xfree;
     ftpScheduleReadControlReply(ftpState, 0);
-    commSetTimeout(fd, Config.Timeout.read, ftpTimeout, ftpState);
 }
 
 /* ====================================================================== */
@@ -1083,6 +1108,7 @@ ftpWriteCommand(const char *buf, FtpStateData * ftpState)
 {
     debug(9, 5) ("ftpWriteCommand: %s\n", buf);
     safe_free(ftpState->ctrl.last_command);
+    safe_free(ftpState->ctrl.last_reply);
     ftpState->ctrl.last_command = xstrdup(buf);
     comm_write(ftpState->ctrl.fd,
 	xstrdup(buf),
@@ -1097,8 +1123,6 @@ static void
 ftpWriteCommandCallback(int fd, char *bufnotused, size_t size, int errflag, void *data)
 {
     FtpStateData *ftpState = data;
-    StoreEntry *entry = ftpState->entry;
-    ErrorState *err;
     debug(9, 7) ("ftpWriteCommandCallback: wrote %d bytes\n", size);
     if (size > 0) {
 	fd_bytes(fd, size, FD_WRITE);
@@ -1109,13 +1133,9 @@ ftpWriteCommandCallback(int fd, char *bufnotused, size_t size, int errflag, void
 	return;
     if (errflag) {
 	debug(50, 1) ("ftpWriteCommandCallback: FD %d: %s\n", fd, xstrerror());
-	if (entry->mem_obj->inmem_hi == 0) {
-	    err = errorCon(ERR_WRITE_ERROR, HTTP_SERVICE_UNAVAILABLE);
-	    err->xerrno = errno;
-	    err->request = requestLink(ftpState->request);
-	    errorAppendEntry(entry, err);
-	}
-	comm_close(ftpState->ctrl.fd);
+	ftpFailed(ftpState, ERR_WRITE_ERROR);
+	/* ftpFailed closes ctrl.fd and frees ftpState */
+	return;
     }
 }
 
@@ -1197,6 +1217,14 @@ ftpScheduleReadControlReply(FtpStateData * ftpState, int buffered_ok)
 	    ftpReadControlReply,
 	    ftpState,
 	    Config.Timeout.read);
+	/*
+	 * Cancel the timeout on the Data socket (if any) and
+	 * establish one on the control socket.
+	 */
+	if (ftpState->data.fd > -1)
+	    commSetTimeout(ftpState->data.fd, -1, NULL, NULL);
+	commSetTimeout(ftpState->ctrl.fd, Config.Timeout.read, ftpTimeout,
+	    ftpState);
     }
 }
 
@@ -1206,8 +1234,11 @@ ftpReadControlReply(int fd, void *data)
     FtpStateData *ftpState = data;
     StoreEntry *entry = ftpState->entry;
     int len;
-    ErrorState *err;
     debug(9, 5) ("ftpReadControlReply\n");
+    if (EBIT_TEST(entry->flags, ENTRY_ABORTED)) {
+	comm_close(ftpState->ctrl.fd);
+	return;
+    }
     assert(ftpState->ctrl.offset < ftpState->ctrl.size);
     Counter.syscalls.sock.reads++;
     len = read(fd,
@@ -1224,26 +1255,17 @@ ftpReadControlReply(int fd, void *data)
 	if (ignoreErrno(errno)) {
 	    ftpScheduleReadControlReply(ftpState, 0);
 	} else {
-	    if (entry->mem_obj->inmem_hi == 0) {
-		err = errorCon(ERR_READ_ERROR, HTTP_INTERNAL_SERVER_ERROR);
-		err->xerrno = errno;
-		err->request = requestLink(ftpState->request);
-		errorAppendEntry(entry, err);
-	    }
-	    comm_close(ftpState->ctrl.fd);
+	    ftpFailed(ftpState, ERR_READ_ERROR);
+	    /* ftpFailed closes ctrl.fd and frees ftpState */
+	    return;
 	}
 	return;
     }
     if (len == 0) {
 	if (entry->store_status == STORE_PENDING) {
-	    storeReleaseRequest(entry);
-	    if (entry->mem_obj->inmem_hi == 0) {
-		err = errorCon(ERR_FTP_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
-		err->xerrno = 0;
-		err->request = requestLink(ftpState->request);
-		err->ftp_server_msg = ftpState->ctrl.message;
-		errorAppendEntry(entry, err);
-	    }
+	    ftpFailed(ftpState, ERR_FTP_FAILURE);
+	    /* ftpFailed closes ctrl.fd and frees ftpState */
+	    return;
 	}
 	comm_close(ftpState->ctrl.fd);
 	return;
@@ -1290,7 +1312,7 @@ ftpHandleControlReply(FtpStateData * ftpState)
     safe_free(ftpState->ctrl.last_reply);
     ftpState->ctrl.last_reply = (*W)->key;
     safe_free(*W);
-    debug(9, 8) ("ftpReadControlReply: state=%d, code=%d\n", ftpState->state,
+    debug(9, 8) ("ftpHandleControlReply: state=%d, code=%d\n", ftpState->state,
 	ftpState->ctrl.replycode);
     FTP_SM_FUNCS[ftpState->state] (ftpState);
 }
@@ -1304,6 +1326,8 @@ ftpReadWelcome(FtpStateData * ftpState)
     debug(9, 3) ("ftpReadWelcome\n");
     if (ftpState->flags.pasv_only)
 	ftpState->login_att++;
+    /* Dont retry if the FTP server accepted the connection */
+    ftpState->fwd->flags.dont_retry = 1;
     if (code == 220) {
 	if (ftpState->ctrl.message) {
 	    if (strstr(ftpState->ctrl.message->key, "NetWare"))
@@ -1364,13 +1388,13 @@ ftpReadPass(FtpStateData * ftpState)
 {
     int code = ftpState->ctrl.replycode;
     debug(9, 3) ("ftpReadPass\n");
-    if (ftpState->ctrl.message) {
-	if (ftpState->cwd_message)
-	    wordlistDestroy(&ftpState->cwd_message);
-	ftpState->cwd_message = ftpState->ctrl.message;
-	ftpState->ctrl.message = NULL;
-    }
     if (code == 230) {
+	if (ftpState->ctrl.message) {
+	    if (ftpState->cwd_message)
+		wordlistDestroy(&ftpState->cwd_message);
+	    ftpState->cwd_message = ftpState->ctrl.message;
+	    ftpState->ctrl.message = NULL;
+	}
 	ftpSendType(ftpState);
     } else {
 	ftpFail(ftpState);
@@ -1669,6 +1693,11 @@ ftpSendPasv(FtpStateData * ftpState)
     snprintf(cbuf, 1024, "PASV\r\n");
     ftpWriteCommand(cbuf, ftpState);
     ftpState->state = SENT_PASV;
+    /*
+     * ugly hack for ftp servers like ftp.netscape.com that sometimes
+     * dont acknowledge PORT commands.
+     */
+    commSetTimeout(ftpState->data.fd, 15, ftpTimeout, ftpState);
 }
 
 static void
@@ -1685,8 +1714,6 @@ ftpReadPasv(FtpStateData * ftpState)
     debug(9, 3) ("This is ftpReadPasv\n");
     if (code != 227) {
 	debug(9, 3) ("PASV not supported by remote end\n");
-	comm_close(ftpState->data.fd);
-	ftpState->data.fd = -1;
 	ftpSendPort(ftpState);
 	return;
     }
@@ -1721,6 +1748,9 @@ ftpReadPasv(FtpStateData * ftpState)
     debug(9, 5) ("ftpReadPasv: connecting to %s, port %d\n", junk, port);
     ftpState->data.port = port;
     ftpState->data.host = xstrdup(junk);
+    safe_free(ftpState->ctrl.last_command);
+    safe_free(ftpState->ctrl.last_reply);
+    ftpState->ctrl.last_command = xstrdup("Connect to server data port");
     commConnectStart(fd, junk, port, ftpPasvCallback, ftpState);
 }
 
@@ -1728,17 +1758,13 @@ static void
 ftpPasvCallback(int fd, int status, void *data)
 {
     FtpStateData *ftpState = data;
-    request_t *request = ftpState->request;
-    ErrorState *err;
     debug(9, 3) ("ftpPasvCallback\n");
     if (status != COMM_OK) {
-	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
-	err->xerrno = errno;
-	err->host = xstrdup(ftpState->data.host);
-	err->port = ftpState->data.port;
-	err->request = requestLink(request);
-	errorAppendEntry(ftpState->entry, err);
-	comm_close(ftpState->ctrl.fd);
+	debug(9, 2) ("ftpPasvCallback: failed to connect. Retrying without PASV.\n");
+	ftpState->fwd->flags.dont_retry = 0;	/* this is a retryable error */
+	ftpState->fwd->flags.ftp_pasv_failed = 1;
+	ftpFailed(ftpState, ERR_NONE);
+	/* ftpFailed closes ctrl.fd and frees ftpState */
 	return;
     }
     ftpRestOrList(ftpState);
@@ -1752,6 +1778,14 @@ ftpOpenListenSocket(FtpStateData * ftpState, int fallback)
     socklen_t addr_len;
     int on = 1;
     u_short port = 0;
+    /*
+     * * Tear down any old data connection if any. We are about to
+     * * establish a new one.
+     */
+    if (ftpState->data.fd > 0) {
+	comm_close(ftpState->data.fd);
+	ftpState->data.fd = -1;
+    }
     /*
      * Set up a listen socket on the same local address as the
      * control connection.
@@ -1826,8 +1860,6 @@ ftpReadPort(FtpStateData * ftpState)
     if (code != 200) {
 	/* Fall back on using the same port as the control connection */
 	debug(9, 3) ("PORT not supported by remote end\n");
-	comm_close(ftpState->data.fd);
-	ftpState->data.fd = -1;
 	ftpOpenListenSocket(ftpState, 1);
     }
     ftpRestOrList(ftpState);
@@ -1838,10 +1870,10 @@ static void
 ftpAcceptDataConnection(int fd, void *data)
 {
     FtpStateData *ftpState = data;
-    struct sockaddr_in peer, me;
+    struct sockaddr_in my_peer, me;
     debug(9, 3) ("ftpAcceptDataConnection\n");
 
-    fd = comm_accept(fd, &peer, &me);
+    fd = comm_accept(fd, &my_peer, &me);
     if (fd < 0) {
 	debug(9, 1) ("ftpHandleDataAccept: comm_accept(%d): %s", fd, xstrerror());
 	/* XXX Need to set error message */
@@ -1852,8 +1884,8 @@ ftpAcceptDataConnection(int fd, void *data)
     comm_close(ftpState->data.fd);
     debug(9, 3) ("ftpAcceptDataConnection: Connected data socket on FD %d\n", fd);
     ftpState->data.fd = fd;
-    ftpState->data.port = ntohs(peer.sin_port);
-    ftpState->data.host = xstrdup(inet_ntoa(peer.sin_addr));
+    ftpState->data.port = ntohs(my_peer.sin_port);
+    ftpState->data.host = xstrdup(inet_ntoa(my_peer.sin_addr));
     commSetTimeout(ftpState->ctrl.fd, -1, NULL, NULL);
     commSetTimeout(ftpState->data.fd, Config.Timeout.read, ftpTimeout,
 	ftpState);
@@ -1868,7 +1900,6 @@ ftpAcceptDataConnection(int fd, void *data)
 static void
 ftpRestOrList(FtpStateData * ftpState)
 {
-
     debug(9, 3) ("This is ftpRestOrList\n");
     if (ftpState->flags.put) {
 	debug(9, 3) ("ftpRestOrList: Sending STOR request...\n");
@@ -1889,10 +1920,20 @@ ftpRestOrList(FtpStateData * ftpState)
 static void
 ftpSendStor(FtpStateData * ftpState)
 {
-    assert(ftpState->filepath != NULL);
-    snprintf(cbuf, 1024, "STOR %s\r\n", ftpState->filepath);
-    ftpWriteCommand(cbuf, ftpState);
-    ftpState->state = SENT_STOR;
+    if (ftpState->filepath != NULL) {
+	/* Plain file upload */
+	snprintf(cbuf, 1024, "STOR %s\r\n", ftpState->filepath);
+	ftpWriteCommand(cbuf, ftpState);
+	ftpState->state = SENT_STOR;
+    } else if (httpHeaderGetInt(&ftpState->request->header, HDR_CONTENT_LENGTH) > 0) {
+	/* File upload without a filename. use STOU to generate one */
+	snprintf(cbuf, 1024, "STOU\r\n");
+	ftpWriteCommand(cbuf, ftpState);
+	ftpState->state = SENT_STOR;
+    } else {
+	/* No file to transfer. Only create directories if needed */
+	ftpSendReply(ftpState);
+    }
 }
 
 static void
@@ -1900,7 +1941,9 @@ ftpReadStor(FtpStateData * ftpState)
 {
     int code = ftpState->ctrl.replycode;
     debug(9, 3) ("This is ftpReadStor\n");
-    if (code >= 100 && code < 200) {
+    if (code == 125 || (code == 150 && ftpState->data.host)) {
+	/* Begin data transfer */
+	debug(9, 3) ("ftpReadStor: starting data transfer\n");
 	/*
 	 * Cancel the timeout on the Control socket, pumpStart will
 	 * establish one on the data socket.
@@ -1909,15 +1952,17 @@ ftpReadStor(FtpStateData * ftpState)
 	ftpPutStart(ftpState);
 	debug(9, 3) ("ftpReadStor: writing data channel\n");
 	ftpState->state = WRITING_DATA;
-    } else if (code == 553) {
-	/* directory does not exist, have to create, sigh */
-#if WORK_IN_PROGRESS
-	ftpTraverseDirectory(ftpState);
-#endif
-	ftpSendReply(ftpState);
+    } else if (code == 150) {
+	/* Accept data channel */
+	debug(9, 3) ("ftpReadStor: accepting data channel\n");
+	commSetSelect(ftpState->data.fd,
+	    COMM_SELECT_READ,
+	    ftpAcceptDataConnection,
+	    ftpState,
+	    0);
     } else {
-	debug(9, 3) ("ftpReadStor: that's all folks\n");
-	ftpSendReply(ftpState);
+	debug(9, 3) ("ftpReadStor: Unexpected reply code %s\n", code);
+	ftpFail(ftpState);
     }
 }
 
@@ -2099,13 +2144,22 @@ ftpReadTransferDone(FtpStateData * ftpState)
 {
     int code = ftpState->ctrl.replycode;
     debug(9, 3) ("This is ftpReadTransferDone\n");
-    if (code != 226) {
+    if (code == 226) {
+	/* Connection closed; retrieval done. */
+	if (ftpState->flags.html_header_sent)
+	    ftpListingFinish(ftpState);
+	if (!ftpState->flags.put) {
+	    storeTimestampsSet(ftpState->entry);
+	    fwdComplete(ftpState->fwd);
+	}
+	ftpDataTransferDone(ftpState);
+    } else {			/* != 226 */
 	debug(9, 1) ("ftpReadTransferDone: Got code %d after reading data\n",
 	    code);
-	debug(9, 1) ("--> releasing '%s'\n", storeUrl(ftpState->entry));
-	storeReleaseRequest(ftpState->entry);
+	ftpFailed(ftpState, ERR_FTP_FAILURE);
+	/* ftpFailed closes ctrl.fd and frees ftpState */
+	return;
     }
-    ftpDataTransferDone(ftpState);
 }
 
 static void
@@ -2204,7 +2258,6 @@ ftpHackShortcut(FtpStateData * ftpState, FTPSM * nextState)
 static void
 ftpFail(FtpStateData * ftpState)
 {
-    ErrorState *err;
     debug(9, 3) ("ftpFail\n");
     /* Try the / hack to support "Netscape" FTP URL's for retreiving files */
     if (!ftpState->flags.isdir &&	/* Not a directory */
@@ -2235,45 +2288,80 @@ ftpFail(FtpStateData * ftpState)
 	    break;
 	}
     }
+    ftpFailed(ftpState, ERR_NONE);
+    /* ftpFailed closes ctrl.fd and frees ftpState */
+}
+
+static void
+ftpFailed(FtpStateData * ftpState, err_type error)
+{
+    StoreEntry *entry = ftpState->entry;
+    if (entry->mem_obj->inmem_hi == 0)
+	ftpFailedErrorMessage(ftpState, error);
+    if (ftpState->data.fd > -1) {
+	comm_close(ftpState->data.fd);
+	ftpState->data.fd = -1;
+    }
+    comm_close(ftpState->ctrl.fd);
+}
+
+static void
+ftpFailedErrorMessage(FtpStateData * ftpState, err_type error)
+{
+    ErrorState *err;
+    char *command, *reply;
     /* Translate FTP errors into HTTP errors */
     err = NULL;
-    switch (ftpState->state) {
-    case SENT_USER:
-    case SENT_PASS:
-	if (ftpState->ctrl.replycode > 500)
-	    err = errorCon(ERR_FTP_FORBIDDEN, HTTP_FORBIDDEN);
-	else if (ftpState->ctrl.replycode == 421)
-	    err = errorCon(ERR_FTP_UNAVAILABLE, HTTP_SERVICE_UNAVAILABLE);
+    switch (error) {
+    case ERR_NONE:
+	switch (ftpState->state) {
+	case SENT_USER:
+	case SENT_PASS:
+	    if (ftpState->ctrl.replycode > 500)
+		err = errorCon(ERR_FTP_FORBIDDEN, HTTP_FORBIDDEN);
+	    else if (ftpState->ctrl.replycode == 421)
+		err = errorCon(ERR_FTP_UNAVAILABLE, HTTP_SERVICE_UNAVAILABLE);
+	    break;
+	case SENT_CWD:
+	case SENT_RETR:
+	    if (ftpState->ctrl.replycode == 550)
+		err = errorCon(ERR_FTP_NOT_FOUND, HTTP_NOT_FOUND);
+	    break;
+	default:
+	    break;
+	}
 	break;
-    case SENT_CWD:
-    case SENT_RETR:
-	if (ftpState->ctrl.replycode == 550)
-	    err = errorCon(ERR_FTP_NOT_FOUND, HTTP_NOT_FOUND);
+    case ERR_READ_TIMEOUT:
+	err = errorCon(error, HTTP_GATEWAY_TIMEOUT);
 	break;
     default:
+	err = errorCon(error, HTTP_BAD_GATEWAY);
 	break;
     }
     if (err == NULL)
 	err = errorCon(ERR_FTP_FAILURE, HTTP_BAD_GATEWAY);
+    err->xerrno = errno;
     err->request = requestLink(ftpState->request);
-    err->ftp_server_msg = ftpState->ctrl.message;
+    err->ftp.server_msg = ftpState->ctrl.message;
+    ftpState->ctrl.message = NULL;
     if (ftpState->old_request)
-	err->ftp.request = ftpState->old_request;
+	command = ftpState->old_request;
     else
-	err->ftp.request = ftpState->ctrl.last_command;
-    if (err->ftp.request) {
-	if (!strncmp(err->ftp.request, "PASS", 4))
-	    err->ftp.request = "PASS <yourpassword>";
-    }
+	command = ftpState->ctrl.last_command;
+    if (command && strncmp(command, "PASS", 4) == 0)
+	command = "PASS <yourpassword>";
     if (ftpState->old_reply)
-	err->ftp.reply = ftpState->old_reply;
+	reply = ftpState->old_reply;
     else
-	err->ftp.reply = ftpState->ctrl.last_reply;
-    errorAppendEntry(ftpState->entry, err);
-    comm_close(ftpState->ctrl.fd);
+	reply = ftpState->ctrl.last_reply;
+    if (command)
+	err->ftp.request = xstrdup(command);
+    if (reply)
+	err->ftp.reply = xstrdup(reply);
+    fwdFail(ftpState->fwd, err);
 }
 
-void
+static void
 ftpPumpClosedData(int data_fd, void *data)
 {
     FtpStateData *ftpState = data;
@@ -2329,6 +2417,9 @@ ftpSendReply(FtpStateData * ftpState)
     if (code == 226) {
 	err_code = (ftpState->mdtm > 0) ? ERR_FTP_PUT_MODIFIED : ERR_FTP_PUT_CREATED;
 	http_code = (ftpState->mdtm > 0) ? HTTP_ACCEPTED : HTTP_CREATED;
+    } else if (code == 227) {
+	err_code = ERR_FTP_PUT_CREATED;
+	http_code = HTTP_CREATED;
     } else {
 	err_code = ERR_FTP_PUT_ERROR;
 	http_code = HTTP_INTERNAL_SERVER_ERROR;
@@ -2336,16 +2427,16 @@ ftpSendReply(FtpStateData * ftpState)
     err = errorCon(err_code, http_code);
     err->request = requestLink(ftpState->request);
     if (ftpState->old_request)
-	err->ftp.request = ftpState->old_request;
+	err->ftp.request = xstrdup(ftpState->old_request);
     else
-	err->ftp.request = ftpState->ctrl.last_command;
+	err->ftp.request = xstrdup(ftpState->ctrl.last_command);
     if (ftpState->old_reply)
-	err->ftp.reply = ftpState->old_reply;
+	err->ftp.reply = xstrdup(ftpState->old_reply);
     else
-	err->ftp.reply = ftpState->ctrl.last_reply;
+	err->ftp.reply = xstrdup(ftpState->ctrl.last_reply);
     errorAppendEntry(ftpState->entry, err);
     storeBufferFlush(ftpState->entry);
-    comm_close(ftpState->ctrl.fd);
+    ftpSendQuit(ftpState);
 }
 
 static void
